## Utility functions

expand <- function(X)
    ## expand design to include second order terms
{
    ## add columns for constant, X, X^2 and all X_i * X_j
    N <- ncol(X)
    d <- as.data.frame(X)
    M <- model.matrix(~ (.)^2, data = d)
    ans <- cbind(M[, 1:(N+1)], X^2, M[, (N+2):ncol(M)])
    colnames(ans) <- NULL
    ans
}

logM <- function(X)
    ## calculate log(det(M(d)))
{
    ans <- determinant(crossprod(expand(X)) / nrow(X),
                       logarithm = TRUE)$modulus
    attributes(ans) <- NULL
    ans
}

logMopt <- function(k) 
    ## log(det(M(.))) for approximate D-optimal design
    ## k is number of input factors
{
    u <- ((k+3) * ((2*k^2 + 3*k + 7) + (k-1) * sqrt(4*k^2 + 12*k + 17)) /
          (4 * (k+1) * (k+2)^2))
    v <- ((k+3) * ((4*k^3 + 8*k^2 + 11*k - 5) + (2*k^2 + k + 3) * sqrt(4*k^2 + 12 * k + 17)) /
          (8 * (k+2)^3 * (k+1)))
    k * log(u) + 0.5 * k * (k-1) * log(v) + (k-1) * log(u-v) + log(u + (k-1)*v - k*u^2)
}


Deff <- function(X)
    ## D-efficiency of design X
{
    m <- ncol(X)
    p <- (m+2)*(m+1)/2 # ncol(expand(X))
    logDE <- log(100) + (logM(X) - logMopt(m)) / p
    exp(logDE)
}

## Generate random design on the cube.

r3 <- function(n) sample(c(-1, 0, 1), n, replace = TRUE)

randomDesign <- function(k, n)
{
    matrix(r3(k*n), n, k)
}

bestRandomDesign <- function(k = 4, n = 30, tries = 1000)
    ## Choose best among several random designs
{
    bestLogM <- -Inf
    dbest <- NULL
    for (i in seq_len(tries))
    {
        d <- randomDesign(k, n)
        if ((tmp <- logM(d)) > bestLogM)
        {
            bestLogM <- tmp
            dbest <- d
        }
    }
    dbest
}


bestImprovement <- function(d)
    ## find single change that would lead to best improvement.
{
    N <- length(d)
    ## Form Nx3 matrix
    e <- matrix(0, N, 3)
    for (i in 1:N)
        for (j in 1:3)
        {
            r <- j-2 # replacement
            if (d[i] != r) # o.w. 0
            {
                nd <- d
                nd[i] <- j-2
                e[i,j] <- Deff(nd)
            }
        }
    if (max(e) > Deff(d))
    {
        w <- which.max(e)
        list(i = row(e)[w], r = col(e)[w]-2)
    }
    else
        NULL
}


improveDesignRandom <- function(d, niter = 1000)
    ## Improve input design by changing elements at random
{
    N <- length(d)
    bestLogM <- logM(d)
    for (i in seq_len(niter))
    {
        nd <- d
        ## choose a single value and change it
        w <- sample(N, 1)
        x <- d[w]
        while (x == d[w]) x <- r3(1)
        nd[w] <- x
        nlogM <- logM(nd)
        if (nlogM > bestLogM) # update d
        {
            bestLogM <- nlogM
            d <- nd
        }
        ## at regular intervals, check if any improvements are at all
        ## possible, and stop if not.
        if (i %% 100 == 0 && is.null(bestImprovement(d)))
        {
            ## message("stopping after ", i, " iterations")
            return(d)
        }
    }
    d
}

improveDesignGreedy <- function(d, niter = 1000)
    ## Improve input design by making change with highest improvement
{
    N <- length(d)
    bestLogM <- logM(d)
    for (i in seq_len(niter))
    {
        best.update <- bestImprovement(d)
        if (is.null(best.update))
        {
            ## message("stopping after ", i, " steps")
            return(d)
        }
        d[best.update$i] <- best.update$r
        ## message(Deff(d))
    }
    d
}


## User functions

findBestDesignRandom <- function(k = 4, n = 30, init.tries = 1000, niter.greedy = 1000)
{
    dbest <- bestRandomDesign(k, n, tries = init.tries)
    ## further random optimize
    dbest <- improveDesignRandom(dbest, niter = niter.greedy)
    message(sprintf("D-efficiency (%d,%d): %g", k, n, round(Deff(dbest), 2)))
    dbest
}

findBestDesignGreedy <- function(k = 4, n = 30, init.tries = 1000, niter.greedy = 1000)
{
    dbest <- bestRandomDesign(k, n, tries = init.tries)
    ## further greedy optimize
    dbest <- improveDesignGreedy(dbest, niter = niter.greedy)
    message(sprintf("D-efficiency (%d,%d): %g", k, n, round(Deff(dbest), 2)))
    dbest
}


bestOutOfN <- function(N, ..., FUN = findBestDesignRandom)
{
    out <- vector(mode = "list", length = N)
    for (I in seq_len(N))
        out[[I]] <- FUN(...)
    eff <- sapply(out, Deff)
    w <- which.max(eff)
    out[[w]]
}


## Example: Obtain design with k=6, N=40.

## Try 5 times and choose best design, with each attempt
## - choosing initial design as best out of 5000 random designs
## - trying 20000 iterations to improve current design

## Method Random.Greedy
bestOutOfN(5, k = 6, n = 40, 5000, 20000, FUN = findBestDesignGreedy)

## Method Random.Random
bestOutOfN(5, k = 6, n = 40, 5000, 20000, FUN = findBestDesignRandom)


